import os
from instructor import OpenAISchema
from pydantic import Field


class ReadPathFunc(OpenAISchema):
    """
    Given a path to a code file, the function will read the raw content of the code file and return it.
    If the user provides a directory path, the function will return the list of files in the directory.
    """

    file_path: str = Field(
        ...,
        example='/Users/home/desktop/project/read_s3_data.py',
        descriptions="read the content of read_s3_data.py",
    )

    class Config:
        title = "read_path_content"

    @classmethod
    def execute(cls, file_path):
        if os.path.isfile(file_path):
            with open(file_path, 'r') as file:
                return file.read()
        elif os.path.isdir(file_path):
            return os.listdir(file_path)
        else:
            return "Invalid path provided."



import click
import questionary

import agent
from agent.utils import *
from agent.types import Step
from agent.function import Chat, Chain
from agent.model import OpenAIModel
from agent.prompt import pmpt_sys_init
from agent.templates import load_step, get_step_mapping

console = Console()
# avoid the tokenizers parallelism issue
os.environ['TOKENIZERS_PARALLELISM'] = 'false'


def build_config(general: bool = False):
    """
    build_config: build the configuration for Termax.
    Args:
        general: a boolean indicating whether to build the general configuration only.
    :return:
    """
    configuration = Config()
    platform = LLM_TYPE_OPENAI
    api_key = questionary.text("What is your OpenAI API key?").ask()

    general_config = {
        'platform': platform,
        'api_key': api_key,
    }

    configuration.write_section(CONFIG_SEC_GENERAL, general_config)

    if not general:
        platform_config = {
            "model": 'gpt-3.5-turbo',
            'temperature': 0.7,
            'max_tokens': 2000,
            'top_p': 1.0,
            'top_k': 32,
            'stop_sequences': 'None',
            'candidate_count': 1
        }

        configuration.write_section(platform, platform_config)


def load_model():
    """
    load_model: load the model based on the configuration.
    """
    configuration = Config()
    config_dict = configuration.read()
    plat = config_dict['general']['platform']

    model = None
    if plat == LLM_TYPE_OPENAI:
        model = OpenAIModel(
            api_key=config_dict[CONFIG_SEC_GENERAL][CONFIG_SEC_API_KEY],
            model=config_dict[LLM_TYPE_OPENAI].get('model'),
            temperature=float(config_dict[LLM_TYPE_OPENAI]['temperature'])
        )

    return model


@click.group()
@click.version_option(version=agent.__version__)
def cli():
    """
    MLE-Agent: The CLI tool to build machine learning projects.
    """
    pass


@cli.command()
@click.option('--general', '-g', is_flag=True, help="Set up the general configuration for MLE Agent.")
def config(general):
    """
    Set up the global configuration for Termax.
    """
    build_config(general)


@cli.command()
def start():
    """
    start: start the chat with LLM.
    """
    chain = Chain(load_step('data_collection.yml'), load_model())
    chain.start()


@cli.command()
def chat():
    """
    chat: start an interactive chat with LLM to work on your ML project.
    """
    # check the system configuration
    configuration = Config()
    if configuration.read() is None:
        build_config()

    model = load_model()
    console.log("Welcome to MLE-Agent! :sunglasses:")
    if model:
        console.log(f"Model loaded: {model.model_type}, {model.model}")
    console.line()

    if configuration.read().get('project') is None:
        console.log("You have not set up a project yet.")
        console.log("Please create a new project first using 'mle new' command.")
        return

    project_path = configuration.read()['project']['path']
    project_state = read_project_state(os.path.join(project_path, CONFIG_PROJECT_FILE))
    console.log("> Current project:", project_path)

    selected_language = project_state.lang
    current_step = project_state.step
    console.log("> Project language:", selected_language)

    # start the interactive chat
    console.line()
    chat_app = Chat(model)
    # set the initial system prompt
    chat_app.add(role='system', content=pmpt_sys_init(selected_language, load_step(get_step_mapping(current_step))))
    chat_app.start()


@cli.command()
def new():
    """
    new: create a new machine learning project.
    """
    configuration = Config()
    name = questionary.text("What is the name of the project?").ask()
    if not name:
        return

    description = questionary.text("What is the description of this project? (Optional)").ask()
    language = questionary.text("What is the major language for this project?").ask()

    if not language:
        console.log("Please provide a valid language. Aborted.")
        return

    project_path = create_directory(name)
    update_project_state(
        project_path,
        {
            'step': 0,
            'task': 0,
            'name': name,
            'description': description,
            'llm': configuration.read()['general']['platform'],
            'path': project_path,
            'lang': language
        }
    )
    configuration.write_section(
        CONFIG_SEC_PROJECT, {
            'path': project_path
        }
    )


@cli.command()
@click.argument('path', nargs=-1, type=click.Path(exists=True))
def set_project(path):
    """
    project: set the current project.
    :return:
    """
    configuration = Config()
    project_path = " ".join(path)
    project_config_path = os.path.join(project_path, CONFIG_PROJECT_FILE)
    if not os.path.exists(project_config_path):
        console.log("The project path is not valid. Please check if `project.yml` exists and try again.")
        return

    configuration.write_section(CONFIG_SEC_PROJECT, {
        'path': project_path
    })

    console.log(f"> Project set to {project_path}")



DB_NAME = 'database'

CONFIG_SEC_GENERAL = 'general'
CONFIG_SEC_PROJECT = 'project'
CONFIG_SEC_API_KEY = 'api_key'
CONFIG_PROJECT_FILE = 'project.yml'
CONFIG_CHAT_HISTORY_FILE = 'project.history'
CONFIG_TASK_HISTORY_FILE = 'task.history'
CONFIG_LLM_LIST = {  # with the default model.
    'OpenAI': 'gpt-3.5-turbo',
    'Gemini': 'gemini-pro',
    'Claude': 'claude-3-opus-20240229',
    'Mistral': 'mistral-small-latest',
    'Qianfan': 'ERNIE-3.5-8K',
    'Qianwen': 'qwen-turbo'
}

# LLMs
LLM_TYPE_OPENAI = 'openai'

# Integrations
INTEGRATIOIN_AWS_S3 = 'aws_s3'
INTEGRATIOIN_HUGGINGFACE = 'huggingface'
INTEGRATIOIN_SNOWFLAKE = 'snowflake'
INTEGRATION_LIST = [
    INTEGRATIOIN_AWS_S3,
    INTEGRATIOIN_HUGGINGFACE,
    INTEGRATIOIN_SNOWFLAKE
]



import os
import questionary


def pmpt_s3_credentials():
    return f"""
    The credentials for the S3 bucket are stored in the environment variables with the following names:
    
    - Access Key ID: 'AWS_ACCESS_KEY_ID'
    - Secret Access Key: AWS_SECRET_ACCESS_KEY'
    """


def aws_credentials():
    """
    Prompt the user for the AWS credentials and set to the environment variables.
    :return: the AWS Access Key ID and the AWS Secret Access Key.
    """
    aws_access_key_id = questionary.text("Enter the AWS Access Key ID:").ask()
    aws_secret_access_key = questionary.text("Enter the AWS Secret Access Key:").ask()

    os.environ['AWS_ACCESS_KEY_ID'] = aws_access_key_id
    os.environ['AWS_SECRET_ACCESS_KEY'] = aws_secret_access_key
    return aws_access_key_id, aws_secret_access_key



import uuid
import os.path
from datetime import datetime
from typing import List, Dict

from agent.const import *
from agent.utils import Config, CONFIG_HOME

import chromadb
from chromadb.utils import embedding_functions

chromadb.logger.setLevel(chromadb.logging.ERROR)


class Memory:
    def __init__(
            self,
            data_path: str = CONFIG_HOME
    ):
        """
        RAG for Termax: memory and external knowledge management.
        Args:
            data_path: the path to store the data.

        """
        self.config = Config().read()
        self.client = chromadb.PersistentClient(path=os.path.join(data_path, DB_NAME))

    def create_collection(self, collection_name: str, embedding_model: str = "text-embedding-ada-002"):
        """
        create_collection: create a collection in the memory.

        :param collection_name: the name of the collection.
        :param embedding_model: the embedding model to use, default will use the embedding model from ChromaDB,
             if the OpenAI has been set in the configuration, it will use the OpenAI embedding model
             "text-embedding-ada-002".

        :return: the collection.
        """
        # use the OpenAI embedding function if the openai section is set in the configuration.
        if self.config.get(LLM_TYPE_OPENAI, None):
            self.client.get_or_create_collection(
                collection_name,
                embedding_function=embedding_functions.OpenAIEmbeddingFunction(
                    model_name=embedding_model,
                    api_key=self.config[LLM_TYPE_OPENAI][CONFIG_SEC_API_KEY]
                )
            )
        else:
            self.client.get_or_create_collection(collection_name)

    def add_query(
            self,
            queries: List[Dict[str, str]],
            collection: str,
            idx: List[str] = None
    ):
        """
        add_query: add the queries to the memery.
        Args:
            queries: the queries to add to the memery. Should be in the format of
                {
                    "query": "the query",
                    "response": "the response"
                }
            collection: the name of the collection to add the queries.
            idx: the ids of the queries, should be in the same length as the queries.
            If not provided, the ids will be generated by UUID.

        Return: A list of generated IDs.
        """
        if idx:
            ids = idx
        else:
            ids = [str(uuid.uuid4()) for _ in range(len(queries))]

        query_list = [query['query'] for query in queries]
        added_time = datetime.now().isoformat()
        resp_list = [{'response': query['response'], 'created_at': added_time} for query in queries]
        # insert the record into the database
        self.client.get_or_create_collection(collection).add(
            documents=query_list,
            metadatas=resp_list,
            ids=ids
        )

        return ids

    def query(self, query_texts: List[str], collection: str, n_results: int = 5):
        """
        query: query the memery.
        Args:
            query_texts: the query texts to search in the memery.
            collection: the name of the collection to search, default is the command history.
            n_results: the number of results to return.

        Returns: the top k results.
        """
        return self.client.get_or_create_collection(collection).query(query_texts=query_texts, n_results=n_results)

    def peek(self, collection: str, n_results: int = 20):
        """
        peek: peek the memery.
        Args:
            collection: the name of the collection to peek, default is the command history.
            n_results: the number of results to return.

        Returns: the top k results.
        """
        return self.client.get_or_create_collection(collection).peek(limit=n_results)

    def get(self, collection: str, record_id: str = None, ):
        """
        get: get the record by the id.
        Args:
            collection: the name of the collection to get the record.
            record_id: the id of the record.

        Returns: the record.
        """
        collection = self.client.get_collection(collection)
        if not record_id:
            return collection.get()

        return collection.get(record_id)

    def delete(self, collection_name: str):
        """
        delete: delete the memery collections.
        Args:
            collection_name: the name of the collection to delete.
        """
        return self.client.delete_collection(name=collection_name)

    def count(self, collection_name: str):
        """
        count: count the number of records in the memery.
        Args:
            collection_name: the name of the collection to count.
        """

        return self.client.get_collection(name=collection_name).count()

    def reset(self):
        """
        reset: reset the memory.
        Notice: You may need to set the environment variable `ALLOW_RESET` to `TRUE` to enable this function.
        """
        self.client.reset()



from abc import ABC, abstractmethod


class Model(ABC):

    def __init__(self):
        self.model_type = None

    @abstractmethod
    def completions(self, chat_history, use_function):
        pass



from pydantic import BaseModel
from typing import List, Optional


class Resource(BaseModel):
    name: str
    description: Optional[str] = None
    choices: Optional[List[str]] = None


class Function(BaseModel):
    name: str
    description: Optional[str] = None


class Task(BaseModel):
    name: str
    kind: str
    description: str
    resources: Optional[List[Resource]] = None
    functions: Optional[List[str]] = None


class Step(BaseModel):
    step: int
    name: str
    description: Optional[str] = None
    tasks: Optional[List[Task]] = None



from .files import ReadPathFunc


def get_function(name: str):
    """
    Get the function by name.
    :param name: the function name.
    :return: the function.
    """
    if name == "read_path_content":
        return ReadPathFunc
    return None


def get_all_func_schema():
    """
    Get all function schema.
    :return: a list of function schema.
    """
    return [ReadPathFunc.openai_schema]



import os
import json
from rich.live import Live
from rich.panel import Panel
from rich.console import Console
from rich.markdown import Markdown

from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory

from agent.utils import Config
from agent.utils import extract_and_save_file, list_all_files
from agent.integration import get_function
from agent.const import CONFIG_CHAT_HISTORY_FILE

config = Config()


class Chat:
    def __init__(self, llm_agent):
        """
        Initialize the chat class.
        :param llm_agent: the language model agent.
        """
        self.agent = llm_agent
        self.chat_history = []
        self.console = Console()
        self.project_home = config.read().get('project')['path']
        self.session = PromptSession(
            history=FileHistory(str(os.path.join(self.project_home, CONFIG_CHAT_HISTORY_FILE)))
        )

    def add(self, role: str, content: str):
        """
        Add the chat to the history.
        :param role: the role.
        :param content: the content.
        :return: the chat history.
        """
        self.chat_history.append({"role": role, "content": content})
        return self.chat_history

    def handle_function_call(self, name, arguments):
        """
        Handle the function call.
        :param name: the function name.
        :param arguments: the function arguments.
        :return:
        """
        self.chat_history.append(
            {
                "role": "assistant",
                "content": "",
                "function_call": {"name": name, "arguments": arguments},
            }
        )

        if arguments != "":
            dict_args = json.loads(arguments)
            joined_args = ", ".join(f'{k}="{v}"' for k, v in dict_args.items())
        else:
            dict_args = {}
            joined_args = ""
        self.console.log(f"> @FunctionCall `{name}({joined_args})` \n\n")

        results = get_function(name).execute(**dict_args)
        self.chat_history.append({"role": "function", "content": results, "name": name})

        return results

    def handle_response(self, prompt):
        """
        Handle the response from the chat.
        :param prompt: the user prompt.
        :return:
        """
        text = ''
        func_name = func_arguments = ""

        self.chat_history.append({"role": "user", "content": prompt})
        response = self.agent.completions(self.chat_history, stream=True, use_function=True)

        for token in response:
            content = token.choices[0].delta.content
            if content:
                text = text + content
                yield text

            function_call = token.choices[0].delta.function_call
            if function_call:
                if function_call.name:
                    func_name = function_call.name
                if function_call.arguments:
                    func_arguments += function_call.arguments

            stop_reason = token.choices[0].finish_reason
            if stop_reason == "function_call":
                self.handle_function_call(func_name, func_arguments)
                yield from self.handle_response(prompt)

            if stop_reason == "stop":
                saved_file, _ = extract_and_save_file(text)
                if saved_file:
                    self.console.log(f"Generating code to: {saved_file}")

        self.chat_history.append({"role": "assistant", "content": text})

    def handle_streaming(self, prompt, display=True):
        """
        Handle the streaming of the chat.
        :param prompt: the user prompt.
        :param display: the flag to display the chat.
        :return:
        """
        if display:
            with Live(console=self.console) as live:
                for text in self.handle_response(prompt):
                    live.update(
                        Panel(Markdown(text), title="[bold magenta]MLE Assistant[/]", border_style="magenta"),
                        refresh=True
                    )
        else:
            self.handle_response(prompt)

    def start(self):
        """
        Start the chat.
        :return: None
        """
        while True:
            try:
                user_pmpt = self.session.prompt("[type to ask]: ").strip()
                # update the local file structure
                self.add("user", f"""
                    The files under the project directory is: {list_all_files(config.read()['project']['path'])}
                """)
                self.handle_streaming(user_pmpt)
            except (KeyboardInterrupt, EOFError):
                exit()



import os
import yaml
from typing import Type, TypeVar
from pydantic import BaseModel, ValidationError

from agent.types import Step

T = TypeVar('T', bound=BaseModel)


def load_step(file_name: str) -> Step:
    """
    Load a step from a .yaml file.
    :param file_name: the name of the configuration file.
    :return:
    """

    dir_path = os.path.dirname(os.path.realpath(__file__))
    yml_path = os.path.join(dir_path, file_name)

    with open(yml_path, 'r') as file:
        data = yaml.safe_load(file)
    try:
        config = Step(**data)
        return config
    except ValidationError as e:
        print(f"Error in loading step file: {e}")
        raise


def get_step_mapping(step_num: int):
    """
    Get the step mapping based on the step number.
    :param step_num: the step number.
    :return: the step mapping.
    """
    step_mapping = {
        0: 'data_collection.yml',
        1: 'data_engineering.yml',
        2: 'model_selection.yml',
        3: 'model_training.yml',
        4: 'model_evaluation.yml',
        5: 'model_deployment.yml'
    }

    return step_mapping.get(step_num, None)


def load_yml_to_pydantic_model(file_path: str, model: Type[T]) -> T:
    """
    Loads YAML data from a file and converts it into a Pydantic model.

    Args:
    file_path (str): Path to the YAML file.
    model (Type[T]): The Pydantic model class to which the data should be converted.

    Returns:
    T: An instance of the specified Pydantic model class.
    """
    with open(file_path, 'r') as file:
        data = yaml.safe_load(file)
        return model(**data)


from agent.types import Step



def pmpt_sys_init(
        lang: str,
        step: Step
) -> str:
    return f"""
    You are an Machine learning engineer, and you are currently working on an ML project using {lang} as the primary language. The ML project contains multiple steps, and each step is a task that you need to complete by generating a code script and the corresponding file name based on the user requirements.
    
    Now, you are currently on step {step.step}:{step.name}
    The output format should be:
    
    File Name: {{name}}
    
    Code: {{code}}
    
    """


def pmpt_ml_project():
    return f"""
    
    """



import importlib.util

from .base import Model
from agent.integration import get_all_func_schema
from agent.const import LLM_TYPE_OPENAI


class OpenAIModel(Model):
    def __init__(self, api_key, model, temperature):
        """
        Initialize the OpenAI model.
        Args:
            api_key (str): The OpenAI API key.
            model (str): The model with version.
            temperature (float): The temperature value.
        """
        super().__init__()

        dependency = "openai"
        spec = importlib.util.find_spec(dependency)
        if spec is not None:
            self.OpenAI = importlib.import_module(dependency).OpenAI
            self.RateLimitError = importlib.import_module(dependency).RateLimitError
        else:
            raise ImportError(
                "It seems you didn't install openai. In order to enable the OpenAI client related features, "
                "please make sure openai Python package has been installed. "
                "More information, please refer to: https://openai.com/product"
            )

        self.model = model
        self.model_type = LLM_TYPE_OPENAI
        self.temperature = temperature
        self.client = self.OpenAI(api_key=api_key)

    def completions(
            self,
            chat_history,
            use_function=False,
            stream=True
    ):
        """
        Completions of the LLM model.
        Args:
            chat_history: The context (chat history).
            use_function: The flag to use the function.
            stream: The flag to stream the output.
        """

        if use_function:
            return self.client.chat.completions.create(
                model=self.model,
                messages=chat_history,
                temperature=self.temperature,
                functions=get_all_func_schema(),
                stream=stream
            )

        return self.client.chat.completions.create(
            model=self.model,
            messages=chat_history,
            temperature=self.temperature,
            stream=stream
        )



import os
import re
import sys
import yaml
import subprocess
from rich.console import Console

from agent.utils import Config
from agent.types import ProjectState
from agent.const import CONFIG_PROJECT_FILE


def list_all_files(path):
    """
    Lists all files and directories under the given path if it is a directory.
    If the path is a file, returns None.

    Args:
    path (str): The file system path to check and list contents from.

    Returns:
    list of str or None: A list of file and directory names under the given path, or None if the path is a file.
    """
    if os.path.isfile(path):
        return None  # Return None if the path is a file

    file_list = []
    for root, dirs, files in os.walk(path):
        for name in files:
            file_list.append(os.path.join(root, name))
        for name in dirs:
            file_list.append(os.path.join(root, name))

    return file_list


def create_directory(dir_name: str):
    """
    Create a directory if it does not exist.
    :param dir_name: the name of the directory to create.
    :return: the directory created.
    """
    console = Console()
    cwd = os.getcwd()
    path = os.path.join(cwd, dir_name)

    try:
        os.makedirs(path, exist_ok=True)
        console.log(f"[green]Directory '{path}' created successfully.")
    except OSError as error:
        console.log(f"[red]Creation of the directory '{path}' failed due to: {error}")
    return path


def get_directory_name(path):
    """
    Get the directory name if the path is a directory.
    :param path: the path to check.
    :return: the directory name if it's a directory, otherwise None.
    """
    if os.path.isdir(path):
        return os.path.basename(path)
    else:
        return None


def read_project_state(config_path: str = None):
    """
    Read the project state.
    :return: the project state.
    """
    console = Console()
    if not config_path:
        config_path = os.path.join(os.getcwd(), CONFIG_PROJECT_FILE)

    if not os.path.exists(config_path):
        console.log(f"[red]The file {config_path} does not exist.")
        return None

    try:
        with open(config_path, 'r') as file:
            data = yaml.safe_load(file)
            return ProjectState(**data)
    except FileNotFoundError:
        console.log(f"[red]The file {config_path} does not exist.")
        return None
    except yaml.YAMLError as error:
        console.log(f"[red]Error parsing YAML file: {error}")
        return None
    except Exception as error:
        console.log(f"[red]An error occurred: {error}")
        return None


def update_project_state(project_path: str, content_dict: dict = None):
    """
    Update the project state.
    :param project_path: the path of the project.
    :param content_dict: the content dictionary to update.
    """
    console = Console()
    file_path = os.path.join(project_path, CONFIG_PROJECT_FILE)

    try:
        with open(file_path, 'w') as file:
            yaml.dump(content_dict, file)
        console.log(f"[green]File '{file_path}' updated successfully.")
    except IOError as error:
        console.log(f"[red]Updating the file '{file_path}' failed due to: {error}")


def extract_and_save_file(input_text):
    """
    Extracts the file name and code block from a text formatted as specified,
    then creates a local file with the file name and writes the code into it.

    Args:
    input_text (str): The text containing the file name and code block.

    Returns:
    str: The name of the file created.
    """
    console = Console()
    file_name_match = re.search(r"File Name:\s*(.*?)\s*\n+\s*Code:", input_text, re.DOTALL)
    if not file_name_match:
        console.log("File name not found in the text.")
        return None, None

    file_name = file_name_match.group(1).strip()

    project_file_path = os.path.join(Config().read()['project']['path'], file_name)
    code_match = re.search(r"```(?:[a-zA-Z0-9]+)?\n(.*?)```", input_text, re.DOTALL)
    if not code_match:
        console.log("Code block not found in the text.")
        return None, None

    code = code_match.group(1).strip()
    with open(project_file_path, 'w') as file:
        file.write(code)

    return project_file_path, code


def extract_file_name(text: str) -> str:
    """
    Extracts the file name from a given text string.

    Args:
    text (str): The text containing the file name.

    Returns:
    str: The extracted file name, or an empty string if no file name is found.
    """
    match = re.search(r'File Name: (\S+)', text)
    if match:
        return match.group(1)
    else:
        return None


def extract_code(text: str) -> str:
    """
    Extracts the code block from a given text string.

    Args:
    text (str): The text containing the code block.

    Returns:
    str: The extracted code block, or an empty string if no code block is found.
    """
    match = re.search(r'```(?:[a-zA-Z0-9]+)?\n(.*?)```', text, re.DOTALL)
    if match:
        return match.group(1)
    else:
        return None


def load_yaml_file(file_path: str):
    """
    Load a YAML file and return the data.
    :param file_path: the path of the YAML file.
    :return: the data in the YAML file.
    """
    with open(file_path, 'r') as file:
        data = yaml.safe_load(file)
    return data


def read_file_to_string(file_path: str):
    """
    Reads the contents of a file and returns it as a string.

    Args:
    file_path (str): The path to the file that needs to be read.

    Returns:
    str: The contents of the file as a string.
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            return file.read()
    except FileNotFoundError:
        return None


def run_command(command):
    """
    Run a command in the shell and return the output and error.
    :param command: the input command to run.
    :return: the output and error.
    """
    try:
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        output = ''
        while True:
            line = process.stdout.readline()
            if not line and process.poll() is not None:
                break
            output += line

        return output
    except Exception as e:
        return str(e)


from pydantic import BaseModel
from typing import Optional


class ProjectState(BaseModel):
    name: str
    step: int
    task: int
    path: str
    llm: str
    lang: str
    description: Optional[str] = None
    target_file: Optional[str] = None
    user_requirement: Optional[str] = None



import questionary
from rich.live import Live
from rich.panel import Panel
from rich.markdown import Markdown

from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory

from agent.utils import *
from agent.types import Step, Task
from agent.const import CONFIG_TASK_HISTORY_FILE

config = Config()


class Chain:

    def __init__(self, step: Step, llm_agent):
        """
        Chain: the interactive chain of the current ML task.
        :param step: the step of the chain.
        :param llm_agent: the language model agent.
        """
        self.step = step
        self.agent = llm_agent
        self.chat_history = []
        self.console = Console()
        project_path = config.read()['project']['path']
        self.project_state = read_project_state(str(os.path.join(project_path, CONFIG_PROJECT_FILE)))
        self.project_home = config.read().get('project')['path']
        self.session = PromptSession(
            history=FileHistory(str(os.path.join(self.project_home, CONFIG_TASK_HISTORY_FILE)))
        )
        self.target_source = self.project_state.target_file
        self.user_requirement = self.project_state.user_requirement

    def ask_choice(self, task: Task):
        """
        Ask a choice question.
        :param task: the task to work on.
        :return: the answer.
        """
        source_kind = questionary.select(task.description, [c.name for c in task.resources]).ask()
        if source_kind:
            for choice in task.resources:
                if choice.name == source_kind:
                    if choice.choices:
                        return questionary.select("Please select", choice.choices).ask()
                    return source_kind

    def gen_file_name(self, user_requirement: str):
        """
        Generate a file name.
        :return: the file name.
        """
        prompt = f"""
        You are an Machine learning engineer, and you are currently working
         on an ML project using {self.project_state.lang} as the primary language.
        Now are are given a user requirement to generate a file name for the current task,
         note the file suffix (e.g., .py) should be correct.
        
        Output format should be:
        
        File Name: {{name}}
        
        """
        self.user_requirement = user_requirement
        self.chat_history.extend(
            [
                {"role": 'system', "content": prompt},
                {"role": 'user', "content": self.user_requirement}
            ]
        )

        with self.console.status("Generating file name..."):
            completion = self.agent.completions(self.chat_history, stream=False)
            target_name = extract_file_name(completion.choices[0].message.content)
            self.target_source = os.path.join(self.project_state.path, target_name)

        self.console.print(f"The generated file name is: {self.target_source}")
        confirm = questionary.confirm("Do you want to use this name?").ask()
        if not confirm:
            new_name = questionary.text("Please provide a new file name:", default=self.target_source).ask()
            if new_name:
                self.target_source = os.path.join(self.project_state.path, new_name)

        # clear the chat history
        self.project_state.target_file = self.target_source
        self.chat_history = []

        return self.target_source

    def gen_task_content(self, task: Task, params=None):
        """
        Generate the content of the current task.
        :param task: the task to work on.
        :param params: the parameters of the previous task.

        :return: the content of the task.
        """
        sys_prompt = f"""
        You are an Machine learning engineer, and you are currently working on an
         ML project using {self.project_state.lang} as the primary language.
        Please generate a code script for the current task based on following information.
        
        Output format should be:
        
        Code: {{code}}
        """

        if self.project_state.target_file:
            sys_prompt = f"""
            You are an Machine learning engineer, and you are currently working on an ML project using
             {self.project_state.lang} as the primary language.
            Please modify (or add new features to) the following code to meet the task requirements.
            
            Existing Code: {read_file_to_string(self.project_state.target_file)}
            
            The output format should be:
        
            Code: {{code}}
            """

        task_prompt = f"""
        User Requirement: {self.user_requirement}
        Primary language: {self.project_state.lang}
        Current task: {task.name}
        Task description: {task.description}
        """

        if params:
            task_prompt += f"""
            Resources: {params}
            """

        self.chat_history.extend(
            [
                {"role": 'system', "content": sys_prompt},
                {"role": 'user', "content": task_prompt}
            ]
        )

        text = ""
        completion = self.agent.completions(self.chat_history, stream=True)
        with Live(console=self.console) as live:
            for token in completion:
                content = token.choices[0].delta.content
                if content:
                    text = text + content
                    live.update(
                        Panel(Markdown(text), title="[bold magenta]MLE Assistant[/]", border_style="magenta"),
                        refresh=True
                    )

                stop_reason = token.choices[0].finish_reason
                if stop_reason == "stop":
                    code = extract_code(text)
                    if code:
                        with open(self.target_source, 'w') as file:
                            file.write(code)
                        self.console.print(f"Code generated to: {self.target_source}")

    def start(self):
        """
        Execute the chain.
        :return: the result of the chain.
        """
        is_running = True
        while is_running:
            if self.project_state.user_requirement:
                self.console.print(f"User Requirement: {self.project_state.user_requirement}")
            else:
                self.user_requirement = questionary.text("What are the user requirements for the file name?").ask()
                self.project_state.user_requirement = self.user_requirement
            if self.target_source is None:
                if self.user_requirement:
                    self.target_source = self.gen_file_name(self.user_requirement)
                    self.console.print(f"The generated file name is: {self.target_source}")

            # working on the task content.
            task_params = None
            task_num = len(self.step.tasks)
            for task in self.step.tasks:
                if self.project_state.task < task_num:
                    self.console.log(f"Working on task {self.project_state.task + 1}/{task_num}")
                    if task.kind == 'code_generation':
                        self.gen_task_content(task, task_params)
                        task_params = None

                    if task.kind == 'multiple_choice':
                        task_params = self.ask_choice(task)

                    self.project_state.task += 1

            is_running = False
            update_project_state(self.project_home, self.project_state.dict(exclude_none=True))
            if self.project_state.task == task_num:
                self.console.print("Looks like all tasks are completed.")
                return
